@@ -1,5 +1,5 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 2009-2019 Free Software Foundation, Inc.
+   Copyright (C) 2009-2020 Free Software Foundation, Inc.
 
    GNU Mailutils is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -30,44 +30,74 @@
 #include <mailutils/errno.h>
 #include <mailutils/nls.h>
 #include <mailutils/stream.h>
+#include <mailutils/cstr.h>
 #include <mailutils/sys/stream.h>
 
 size_t mu_stream_default_buffer_size = MU_STREAM_DEFBUFSIZ;
 
-#define _stream_event(stream, code, n, p)			\
-  do								\
-    {								\
-      if ((stream)->event_cb &&					\
-          ((stream)->event_mask & _MU_STR_EVMASK(code)))	\
-        (stream)->event_cb (stream, code, n, p);		\
-    }								\
-  while (0)
+static inline void
+_stream_event (mu_stream_t str, int code, unsigned long n, void *p)
+{
+  if (str->event_cb && (str->event_mask & _MU_STR_EVMASK (code)))
+    str->event_cb (str, code, n, p);
+}
 
-#define _bootstrap_event(stream)					    \
-  do									    \
-    {									    \
-      if ((stream)->event_cb &&						    \
-          ((stream)->event_mask & _MU_STR_EVMASK(_MU_STR_EVENT_BOOTSTRAP))) \
-	{								    \
-	  (stream)->event_cb (stream, _MU_STR_EVENT_BOOTSTRAP, 0, NULL);    \
-	  (stream)->event_mask &= ~_MU_STR_EVMASK(_MU_STR_EVENT_BOOTSTRAP); \
-	}								    \
-    }									    \
-  while (0)
-  
+static inline void
+_bootstrap_event (mu_stream_t str)
+{
+  if (str->event_cb &&
+      (str->event_mask & _MU_STR_EVMASK (_MU_STR_EVENT_BOOTSTRAP)))
+    {
+      str->event_cb (str, _MU_STR_EVENT_BOOTSTRAP, 0, NULL);
+      str->event_mask &= ~_MU_STR_EVMASK (_MU_STR_EVENT_BOOTSTRAP);
+    }
+}
 
-#define _stream_stat_incr(s, k, n) \
-  (((s)->statmask & MU_STREAM_STAT_MASK(k)) ? ((s)->statbuf[k] += n) : 0)
+static inline void
+_stream_stat_incr (mu_stream_t s, int k, size_t n)
+{
+  if (s->statmask & MU_STREAM_STAT_MASK (k))
+    s->statbuf[k] += n;
+}
 
-#define _stream_read(str, buf, size, rdbytes) \
-  (_stream_stat_incr ((str), MU_STREAM_STAT_READS, 1),  \
-   (str)->read (str, buf, size, rdbytes))
-#define _stream_write(str, buf, size, wrbytes) \
-  (_stream_stat_incr ((str), MU_STREAM_STAT_WRITES, 1),  \
-   (str)->write (str, buf, size, wrbytes))
-#define _stream_seek(str, pos, poff)			\
-  (_stream_stat_incr ((str), MU_STREAM_STAT_SEEKS, 1),		\
-   (str)->seek (str, pos, poff))
+static inline int
+_stream_read (mu_stream_t str, char *buf, size_t size, size_t *pbytes)
+{
+  int rc;
+  _stream_stat_incr (str, MU_STREAM_STAT_READS, 1);
+  rc = str->read (str, buf, size, pbytes);
+  if (rc == 0)
+    {
+      if (str->statmask & MU_STREAM_STAT_MASK (MU_STREAM_STAT_INLN))
+	str->statbuf[MU_STREAM_STAT_INLN] += mu_mem_c_count (buf, '\n', *pbytes);
+      if (str->statmask & MU_STREAM_STAT_MASK (MU_STREAM_STAT_IN8BIT))
+	str->statbuf[MU_STREAM_STAT_IN8BIT] += mu_mem_8bit_count (buf, *pbytes);
+    }
+  return rc;
+}
+
+static inline int
+_stream_write (mu_stream_t str, char const *buf, size_t size, size_t *pbytes)
+{
+  int rc;
+  _stream_stat_incr (str, MU_STREAM_STAT_WRITES, 1);
+  rc = str->write (str, buf, size, pbytes);
+  if (rc == 0)
+    {
+      if (str->statmask & MU_STREAM_STAT_MASK (MU_STREAM_STAT_OUTLN))
+	str->statbuf[MU_STREAM_STAT_OUTLN] += mu_mem_c_count (buf, '\n', *pbytes);
+      if (str->statmask & MU_STREAM_STAT_MASK (MU_STREAM_STAT_OUT8BIT))
+	str->statbuf[MU_STREAM_STAT_OUT8BIT] += mu_mem_8bit_count (buf, *pbytes);
+    }
+  return rc;
+}
+
+static inline int
+_stream_seek (mu_stream_t str, mu_off_t pos, mu_off_t *poff)
+{
+  _stream_stat_incr (str, MU_STREAM_STAT_SEEKS, 1);
+  return str->seek (str, pos, poff);
+}
 
 static int _stream_read_unbuffered (mu_stream_t stream, void *buf, size_t size,
 				    int full_read, size_t *pnread);
@@ -75,14 +105,14 @@ static int _stream_write_unbuffered (mu_stream_t stream,
 				     const void *buf, size_t size,
 				     int full_write, size_t *pnwritten);
 
-static void
+static inline void
 _stream_setflag (struct _mu_stream *stream, int flag)
 {
   _stream_event (stream, _MU_STR_EVENT_SETFLAG, flag, NULL);
   stream->flags |= flag;
 }
 
-static void
+static inline void
 _stream_clrflag (struct _mu_stream *stream, int flag)
 {
   _stream_event (stream, _MU_STR_EVENT_CLRFLAG, flag, NULL);
@@ -97,6 +127,8 @@ mu_stream_seterr (struct _mu_stream *stream, int code, int perm)
     {
     case 0:
     case EAGAIN:
+    case EINTR:
+    case ENOSYS:
     case EINPROGRESS:
       break;
 
@@ -119,10 +151,6 @@ _mu_stream_seteof (mu_stream_t str)
   _stream_setflag (str, _MU_STR_EOF);
 }
 
-#define _stream_buffer_freespace(s) \
-  ((s)->bufsize - (s)->level)
-#define _stream_buffer_is_full(s) (_stream_buffer_freespace(s) == 0)
-
 #define _stream_curp(s) ((s)->buffer + (s)->pos)
 
 static int
@@ -173,24 +201,6 @@ _stream_fill_buffer (struct _mu_stream *stream)
   return rc;
 }
 
-static int
-_stream_buffer_full_p (struct _mu_stream *stream)
-{
-    switch (stream->buftype)
-      {
-      case mu_buffer_none:
-	break;
-	
-      case mu_buffer_line:
-	return _stream_buffer_is_full (stream)
-	       || memchr (stream->buffer, '\n', stream->level) != NULL;
-
-      case mu_buffer_full:
-	return _stream_buffer_is_full (stream);
-      }
-    return 0;
-}
-
 enum
   {
     FLUSH_WRITE, /* Flush only modified data.  Keep buffer level and position
@@ -635,7 +645,7 @@ _stream_read_unbuffered (mu_stream_t stream, void *buf, size_t size,
 	}
       mu_stream_seterr (stream, rc, rc != 0);
     }
-  if (pnread)
+  if (rc == 0 && pnread)
     *pnread = nread;
   
   return rc;
@@ -975,6 +985,28 @@ mu_stream_getline (mu_stream_t stream, char **pbuf, size_t *psize,
   return mu_stream_getdelim (stream, pbuf, psize, '\n', pread);
 }
 
+/* Return 1 if no more data can be written to the current buffer. */
+static inline int
+_stream_buffer_full_p (struct _mu_stream *stream)
+{
+  /* This function should be called only for buffered streams */
+  if (stream->buftype == mu_buffer_none)
+    return 0;
+
+  if (stream->bufsize == stream->pos)
+    /* No space left in buffer */
+    return 1;
+
+  /* For line buffering, the buffer is flushed immediately after
+     receiving a newline character. */
+  if (stream->buftype == mu_buffer_line &&
+      stream->pos > 0 &&
+      memchr (stream->buffer, '\n', stream->pos) != NULL)
+    return 1;
+
+  return 0;
+}
+
 int
 mu_stream_write (mu_stream_t stream, const void *buf, size_t size,
 		 size_t *pnwritten)
@@ -1008,14 +1040,16 @@ mu_stream_write (mu_stream_t stream, const void *buf, size_t size,
 	{
 	  size_t n;
 	  
-	  if (_stream_buffer_full_p (stream)
-	      && (rc = _stream_flush_buffer (stream, FLUSH_RDWR)))
-	    break;
+	  if (_stream_buffer_full_p (stream))
+	    {
+	      if ((rc = _stream_flush_buffer (stream, FLUSH_RDWR)))
+		break;
+	    }
 
 	  if (size == 0)
 	    break;
 	    
-	  n = _stream_buffer_freespace (stream);
+	  n = stream->bufsize - stream->pos;
 	  if (n > size)
 	    n = size;
 	  memcpy (_stream_curp (stream), bufp, n);
@@ -1195,6 +1229,23 @@ mu_stream_truncate (mu_stream_t stream, mu_off_t size)
       
       if ((rc = _stream_flush_buffer (stream, FLUSH_RDWR)))
 	return rc;
+      if (stream->offset > size)
+	{
+	  stream->offset = size;
+	  stream->level = 0;
+	  stream->pos = 0;
+	}
+      else if (stream->offset + stream->pos > size)
+	{
+	  stream->pos = size - stream->offset;
+	  stream->level = stream->pos;
+	}
+      else if (stream->offset + stream->level > size)
+	{
+	  stream->level = size - stream->offset;
+	  if (stream->pos > stream->level)
+	    stream->pos = stream->level;
+	}
       return stream->truncate (stream, size);
     }
   return ENOSYS;
